import * as Electron from 'electron';
import { registerBrowserWindow } from '../registries/browser-window.registry';
import { Store } from '@tsed/core';
import { ELECTRON_BROWSER_WINDOW } from '../types/electron-browser-window';
import {
  BrowserWindowEventHandlerMetadata,
  BrowserWindowProviderMetadata
} from '../models/browser-window-provider.metadata';

export interface BrowserWindowOptions {
  constructorOptions?: Electron.BrowserWindowConstructorOptions;
  beforeConstructor?: (
    constructorOptions: Electron.BrowserWindowConstructorOptions
  ) => Electron.BrowserWindowConstructorOptions;
}

export function BrowserWindow(options?: BrowserWindowOptions): Function {
  return (target: any): void => {
    registerBrowserWindow({
      provide: target,
      useFactory(app: Electron.App) {
        options = options || {};
        let constructorOptions = options.constructorOptions || {};

        if (!!options.beforeConstructor) {
          constructorOptions = options.beforeConstructor(constructorOptions);
        }

        const window = new Electron.BrowserWindow(constructorOptions);

        const w = new target(app, window);

        const providerMetadata: BrowserWindowProviderMetadata = Store.from(
          w
        ).get(ELECTRON_BROWSER_WINDOW);

        if (!!providerMetadata && !!providerMetadata.handlers) {
          const __this = w;
          const invoke = async (
            instance: any,
            handlerMetadata: BrowserWindowEventHandlerMetadata,
            args: any[]
          ) => {
            const { methodClassName } = handlerMetadata;

            return await instance[methodClassName](...args);
          };

          for (const handler in providerMetadata.handlers) {
            if (providerMetadata.handlers.hasOwnProperty(handler)) {
              const metadata = providerMetadata.handlers[handler];

              window.on(metadata.channel as any, (...args: any[]) => {
                invoke(__this, metadata, args);
              });
            }
          }
        }

        return w;
      }
    });
  };
}
